
import logging,os
import numpy as np


from dataset import save_nc, get_outfile, to_weekly, to_monthly
from analogs import analogs_seas_hcst,analogs_seas_fcst
from tools import cut_domain

def apply_downscale(exp,srcobs,targetobs,target_var,region, conf):
    """
    Wrapper func to apply downscaling methods using xr objects
    saves the resulting .nc accordingly to the configuration.

    Parameters
    ----------
    exp : xarray Dataframe
        df containing the experiment data to be downscaled  
    srcobs : xarray Dataframe
        df containing the source observational data 
        used as predictand during training  
    targetobs : xarray Dataframe
        df containing the target observational data 
        used during training  
    target_var : str
        Variable to be downscaled  
    region : str
        Region where the downscaling is applied
        (for exp subsetting)  
    Conf: Config 
        General recipe config 

    """   

    outdir,outfile = get_outfile(exp,conf.outdir)
    sys = conf.arch.get_system(conf.exp_sys,conf.temp_freq)

    if not os.path.isfile(outdir+outfile) or conf.overwrite:
        
        # interpolation of exp grid to srcobs grid
        exp = exp.interp_like(srcobs[target_var].isel(time=0, drop=True), method="nearest")

        logging.info('cutting domain and loading exp data in memory...')    
        exp = cut_domain(exp,conf.regions[region]).load()

        logging.info('exp load completed...')    

        logging.info('Applying analogs method')
        if exp.system_type == "hcst":
            downs = analogs_seas_hcst(exp[target_var],
                         srcobs[target_var],
                         targetobs[target_var],
                         )
        else:
            downs = analogs_seas_fcst(exp[target_var],
                                      srcobs[target_var],
                                      targetobs[target_var])

        logging.info('exporting to .nc')
        downs.attrs = exp.attrs
        downs["time_bnds"] = exp.time_bnds

        # Removes originally missing sdates before exporting
        # TODO: try applying this step at an earlier stage
        # (Nas are difficult to handle with xarray methods) 
        if downs.system == "cfsv2" and downs.system_type == "hcst":
           downs = downs.where(downs.sdate.isin(downs.sdate[downs.not_missing_sdates]))
        
        save_nc(downs,conf.outdir)
        if sys['type'] == "seasonal":
            save_nc(to_monthly(downs),conf.outdir)
        elif sys['type'] == "subseasonal":
            save_nc(to_weekly(downs),conf.outdir)
            
    else:
        logging.warning(f'file {outdir+outfile} already exists, skipping computation')