# module load R/3.6.2-foss-2015a-bare
library(startR)
library(lubridate)
library(s2dv)
library(multiApply)
library(ClimProjDiags) # for SelBox
source("~/Repos/WindPower/downscaling/analogs.R")

###################################
# Training and test period configuration
###################################
all.years <- 1979:2018 # common period ERA5/MSWEP/MESCAN-SURFEX
holdout.years <- c(1983, 1989, 1990, 1996, 1999, 2003, 2005, 2011) # years hold out for final testing (i.e. meteorological verification)
train.years <- all.years[!all.years %in% holdout.years] # Years for training & hypertunning

holdout.sdates <- ymd(paste0(rep(holdout.years,each=12),sprintf("%02d",1:12),"01"))
train.sdates <- ymd(paste0(rep(train.years,each=12),sprintf("%02d",1:12),"01"))

train.idx <- unlist(lapply(days_in_month(train.sdates),function(x) c(rep(T,x),rep(F,31-x))))
holdout.idx <- unlist(lapply(days_in_month(holdout.sdates),function(x) c(rep(T,x),rep(F,31-x))))

###################################
# Dataset configuration
###################################
datasets <- list(
	era5		= '/esarchive/recon/ecmwf/era5/daily_mean/$var$_f1h/$var$_$sdate$.nc',
	era5.cds	= '/esarchive/recon/ecmwf/era5/daily_mean/$var$_f1h-r1440x721cds/$var$_$sdate$.nc',
	mswep		= '/esarchive/recon/gloh2o/mswep-v2.6/daily_mean/$var$_s0-3h/$var$_$sdate$.nc',
	mescan		= '/esarchive/recon/ecmwf/uerra_mescan/original_files/$var$/daily/$var$_$sdate$.nc',
	mescan.remapbil = '/esarchive/recon/ecmwf/uerra_mescan/daily_mean/$var$_f6h/$var$_$sdate$.nc'
)
	
###################################
# Region configuration
###################################
regions <- list(
	vitigeoss.cat = list(lonmin = 0, lonmax = 3.4,
			      latmin = 40.2, latmax = 43.2),
	vitigeoss.douro = list(lonmin = -9, lonmax = -6,
				latmin = 40.2, latmax = 43.2),
	vitigeoss.campania = list(lonmin = 13.5, lonmax = 16.5,
				   latmin = 39.8, latmax = 42.3),
	iberian.peninsula = list(lonmin = -10, lonmax = 4, 
				  latmin = 35.6, latmax = 44),
	southwest.europe = list(lonmin = -30, lonmax = 20, 
				 latmin = 30, latmax = 50),
	euroatlantic = list(lonmin = -80, lonmax = 30, 
			     latmin = 25, latmax = 60)
)

###################################
# Region/var Selection
###################################
LargeScale <- list(var = 'tas',
		   dataset = datasets$era5.cds,
		   region = regions$southwest.europe)
LocalScale <- list(var = 'tas',
		   dataset = datasets$mescan.remapbil,
		   region = regions$iberian.peninsula)

# LargeScale <- list(var = 'prlr',
#		   dataset = datasets$era5.cds,
#		   region = regions$southwest.europe)
# LocalScale <- list(var = 'prlr',
#		   dataset = datasets$mswep,
#		   region = regions$iberian.peninsula)

LoadObsData <- function(what, sdates, subset.idx) {
	tmp <- Start(
		sdate = format(sdates,"%Y%m"),
		time = 'all',
		longitude = values(list(what$region$lonmin, what$region$lonmax)),
		latitude = values(list(what$region$latmin, what$region$latmax)),
		longitude_reorder = CircularSort(0, 360),
		latitude_reorder = Sort(),
		longitude_var = "longitude",
		latitude_var = "latitude",
		var = what$var,
		dat = what$dataset,
		pattern_dims = 'dat',
		return_vars=list(time = NULL, latitude = NULL, longitude = NULL),
		num_procs = 4, retrieve = T,
		synonims = list(longitude = c('lon', 'longitude'), latitude=c('lat', 'latitude')),
		time_across = 'sdate',
		merge_across_dims = T
	)

	tmp <- s2dverification::Subset(tmp, along = "time", subset.idx, drop=T)
	return(tmp)
}


# Load data sets (X=large scale predictors; y= local scale predictand)
X.train <- LoadObsData(LargeScale, train.sdates, train.idx)
y.train <- LoadObsData(LocalScale, train.sdates, train.idx)
X.holdout <- LoadObsData(LargeScale, holdout.sdates, holdout.idx)
y.holdout <- LoadObsData(LocalScale, holdout.sdates, holdout.idx)


#for (v in c("psl", "g500", "prlr", "sfcWind", "uas", "vas")) {
for (v in c("psl", "g500", "sfcWind", "uas", "vas")) {
	LargeScale <- list(var =v,
			   dataset = datasets$era5.cds,
			   region = regions$southwest.europe)

	X.train <- LoadObsData(LargeScale, train.sdates, train.idx)
	saveRDS(X.train,paste0(v,"_era5.cds_southwest.europe_train.RDS"))
	rm(X.train)
	gc()

	#X.holdout <- LoadObsData(LargeScale, holdout.sdates, holdout.idx)
	#saveRDS(X.holdout,paste0(v,"_era5.cds_southwest.europe_holdout.RDS"))
	#rm(X.holdout)
	#gc()
}

# saveRDS(X.train, 
#
#
#

# X.train <- readRDS("tas_era5.cds_southwest.europe_train.RDS")
# X.holdout <- readRDS("tas_era5.cds_southwest.europe_holdout.RDS")
# y.train <- readRDS("tas_mescan.remapbil_iberian.peninsula_train.RDS")
# y.holdout <- readRDS("tas_mescan.remapbil_iberian.peninsula_holdout.RDS")

lons <- attributes(X.train)$Variables$common$lon
lats <- attributes(X.train)$Variables$common$lat
X.train.subset <- SelBox(X.train, as.vector(lons), as.vector(lats), unlist(regions$iberian.peninsula))$data
X.train.subset.df <- as.data.frame(aperm(Apply(X.train.subset, margins = "time", fun=function(x) { as.vector(x) })$output1 ))

X.holdout.subset <- SelBox(X.holdout, as.vector(lons), as.vector(lats), unlist(regions$iberian.peninsula))$data
X.holdout.subset.df <- as.data.frame(aperm(Apply(X.holdout.subset, margins = "time", fun=function(x) { as.vector(x) })$output1 ))



# Train (i.e. estimate model parameters and tune hyperparameters). Use 5-fold CV here if needed.
model <- train(lcl.train, lrg.train)

# Predict
fcst <- predict(model, lrg.holdout)

an <- analogs(X.train, X.holdout, y.train, k=5, fun=mean)

# Verify
verif <- verify(fcst, lcl.holdout)

train <- slp[10:3683,,]
test <- slp[1:9,,]
obs <- pr[10:3683,,]
verif <- pr[1:9,,]
anal <- analogs(train, test, obs, k = 5, fun = mean)

PlotLayout(PlotEquiMap,c(2,3),abind::abind(slp[3651,,],slp[nn.ind$nn.index[1,],,],along=1), lon=lons, lat=lats, filled.continents=F)
PlotLayout(PlotEquiMap,c(2,3),abind::abind(slp[3652,,],slp[nn.ind$nn.index[2,],,],along=1), lon=lons, lat=lats, filled.continents=F)
PlotLayout(PlotEquiMap,c(2,3),abind::abind(slp[3653,,],slp[nn.ind$nn.index[3,],,],along=1), lon=lons, lat=lats, filled.continents=F)

kk <- SelBox(slp,as.vector(lons),as.vector(lats), c(-15,20,30,50), 2,3)

slp3 <- kk$data
lons3 <- kk$lon
lats3 <- kk$lat
PlotLayout(PlotEquiMap,c(2,3),abind::abind(slp3[3651,,],slp3[nn.ind$nn.index[1,],,],along=1), lon=lons3, lat=lats3, filled.continents=F)
PlotLayout(PlotEquiMap,c(2,3),abind::abind(slp3[3652,,],slp3[nn.ind$nn.index[2,],,],along=1), lon=lons3, lat=lats3, filled.continents=F)
PlotLayout(PlotEquiMap,c(2,3),abind::abind(slp3[3653,,],slp3[nn.ind$nn.index[3,],,],along=1), lon=lons3, lat=lats3, filled.continents=F)

