import xarray as xr
import numpy as np
import pickle
import itertools
from tools import fix_coords, cut_domain, scale, rms_error 
from transformers import VarScaler, VarSelector, VarSubset, VarWeight, AddDayOfYear
from sklearn.pipeline import Pipeline
#from sklearn.preprocessing import StandardScaler
from sklearn.neighbors import KNeighborsRegressor
from sklearn.metrics import make_scorer
from sklearn.model_selection import GridSearchCV

#-----------------------------
# Define predictors and predictands
#-----------------------------
era5_dir = '/esarchive/oper/S2S-downscale/era5/'
uerra_dir = '/esarchive/oper/S2S-downscale/uerra/'
mswep_dir = '/esarchive/oper/S2S-downscale/mswep/'
target_var = 'prlr'
src_vars = {"tas": ['tas','uas','vas','psl','g500','sfcWind'],
            "prlr": ['prlr','uas','vas','psl','g500','sfcWind',
                      'hur850','hus850','ua850','va850']}

if target_var == "tas":
    predictand_dir = uerra_dir
else:
    predictand_dir = mswep_dir

#-----------------------------
# Load predictors
#-----------------------------
files = [f'{era5_dir}/{var}_1979-2018.nc'
         for var in src_vars[target_var]]
era5 = xr.open_mfdataset(files,
                         preprocess = fix_coords,
                         combine="by_coords",
                         chunks = {'lon': 4, 'lat': 4},
                         join = 'left')
era5 = era5.drop("time_bnds")

#-----------------------------
# Load predictand
#-----------------------------
files = f'{predictand_dir}/{target_var}_1979-2018.nc'
uerra = xr.open_dataset(files,
                       chunks = {'lon': 4, 'lat': 4},
                       decode_times = True)
try:
    uerra = uerra.drop("time_bnds")
uerra = fix_coords(uerra)

#-----------------------------
# Plots
#-----------------------------
#import matplotlib.pyplot as plt
#era5['tas'].mean(dim='time').plot()
#uerra['tas'].mean(dim='time').plot()
#plt.show()

#-----------------------------
# Split train test sample based on year in holdout
#-----------------------------
holdout_years = [1979, 1983, 1989, 1990, 1996, 1999, 2003, 2005, 2011]
idx_test = np.isin(era5['time'].dt.year, holdout_years)
idx_train = np.logical_not(idx_test)

#-----------------------------
# Subset the domain/time
# NOTE: We assume that uerra.time == era5.time !!!
#-----------------------------
#X_train = era5.sel(time=idx_train, lon=slice(-10,4), lat=slice(35,44))
X_train = era5.sel(time=idx_train, lon=slice(-2,4), lat=slice(38,44))
X_test = era5.sel(time=idx_test, lon=slice(-2,4), lat=slice(38,44))
y_train = uerra.sel(time=idx_train, lon=slice(0,3.4), lat=slice(40.2,43.2))
y_test = uerra.sel(time=idx_test, lon=slice(0,3.4), lat=slice(40.2,43.2))

#-----------------------------
# Stack multidimensional arrays for sklearn
#-----------------------------
X_train = X_train.to_array().stack(feature=("variable","lon","lat"))
X_test = X_test.to_array().stack(feature=("variable","lon","lat"))
y_train = y_train.prlr.stack(pos=("lon","lat"))
y_test = y_test.prlr.stack(pos=("lon","lat"))

#-----------------------------
# read the data only after subsetting!
#-----------------------------
X_train = X_train.load()
#X_test = X_test.load()
y_train = y_train.load()
#y_test = y_test.load()

#-----------------------------
# Define the field-RMSE metric
#-----------------------------
neg_rmse = make_scorer(rms_error, greater_is_better=False)
all_vars = src_vars[target_var]
#all_vars = np.unique(X_train.coords['variable']).tolist()

###########################################
####### Select best variable combination:
###########################################
pipeline = Pipeline(
    [
        ("scale", VarScaler()),
        ("select", VarSelector(varlist=all_vars)),
        ("knn", KNeighborsRegressor(n_neighbors=1))
    ]
)

params = dict(select__varlist = [["prlr"], ["prlr", "sfcWind"], ["g500","prlr"], ["g500"], 
                                 ["g500","psl"], ["prlr", "psl"], ["prlr", "psl", "g500"],
                                 ["g500","psl","hur850","hus850","va850","ua850"], all_vars],
              knn__n_neighbors = [1, 3, 7, 15, 41])
grid_search = GridSearchCV(pipeline, param_grid=params, scoring=neg_rmse, cv=5, n_jobs=4)
a = grid_search.fit(X_train,y_train)
pickle.dump(a.cv_results_, open('cv_results.v2.pickle','wb'))

###########################################
####### Select best region for tas:
###########################################
pipeline2 = Pipeline(
    [
        ("scale", VarScaler()),
        ("select", VarSelector(varlist="tas")),
        ("subset", VarSubset(var="tas", region=None)),
        ("knn", KNeighborsRegressor(n_neighbors=1))
    ]
)

params = dict(subset__region = [[0,3.3,40.2,43.2], [-0.5,3.8,39.7,43.7], [-1,4,39,44], [-2,4,38,44]],
              knn__n_neighbors = [1, 3, 7, 15, 41])
grid_search2 = GridSearchCV(pipeline2, param_grid=params, scoring=neg_rmse, cv=5, n_jobs=4)
a = grid_search2.fit(X_train,y_train)
pickle.dump(a.cv_results_, open('cv_results.tasregion.pickle','wb'))

###########################################
####### Select best region for g500:
###########################################
pipeline3 = Pipeline(
    [
        ("scale", VarScaler()),
        ("select", VarSelector(varlist=["tas","g500"])),
        ("subset_tas", VarSubset(var="tas", region=[-0.5,3.8,39.7,43.7])),
        ("subset_g500", VarSubset(var="g500", region=None)),
        ("knn", KNeighborsRegressor(n_neighbors=1))
    ]
)

params = dict(subset_g500__region = [[0,3.3,40.2,43.2], [-0.5,3.8,39.7,43.7], [-1,4,39,44], [-2,4,38,44]],
              knn__n_neighbors = [7, 11, 15, 19, 23, 27, 41])
grid_search3 = GridSearchCV(pipeline3, param_grid=params, scoring=neg_rmse, cv=5, n_jobs=4)
a = grid_search3.fit(X_train,y_train)
pickle.dump(a.cv_results_, open('cv_results.g500region.pickle','wb'))

###########################################
####### Select vars again w k and region fixed:
###########################################
pipeline4 = Pipeline(
    [
        ("scale", VarScaler()),
        ("select", VarSelector(varlist=["tas"])),
        ("subset_tas", VarSubset(var="tas", region=[-0.5,3.8,39.7,43.7])), # optimal for tas
        #("subset_other", VarSubset(var=None, region=None)),
        ("knn", KNeighborsRegressor(n_neighbors=15)) # optimal in all experiments so far
    ]
)

comb1vars = [list(set) for set in list(itertools.combinations(all_vars,1))]
comb2vars = [list(set) for set in list(itertools.combinations(all_vars,2))]
comb3vars = [list(set) for set in list(itertools.combinations(all_vars,3))]

params = dict(select__varlist = comb1vars + comb2vars + comb3vars)
grid_search4 = GridSearchCV(pipeline4, param_grid=params, scoring=neg_rmse, cv=5, n_jobs=4)
a = grid_search4.fit(X_train,y_train)
pickle.dump(a.cv_results_, open('cv_results.comb3vars.pickle','wb'))

###########################################
####### g500 with time
###########################################
pipeline5 = Pipeline(
    [
        ("scale", VarScaler()),
        ("select", VarSelector(varlist=['g500'])),
        ("subset_g500", VarSubset(var="g500", region=[-0.5,3.8,39.7,43.7])), # optimal for tas
        ("weight", VarWeight(var="g500", weight=1)),
        ("adddoy", AddDayOfYear(weight=10)),
        ("knn", KNeighborsRegressor(n_neighbors=15)) # optimal in all experiments so far
    ]
)

params = dict(subset_g500__region = [[-1,4,39,44], [-2,4,38,44], [-5,4,36, 44], [-10, 4, 35,44]],
              adddoy__weight = [0,0.008,0.009, 0.01,0.02,0.03,0.05,0.5,1,5,10,20])
#params = dict(subset_g500__region = [[-1,4,39,44], [-2,4,38,44]],
#              adddoy__weight = [0,0.0005,0.001,0.005,0.007,0.008,0.009, 0.01,0.02,0.03,0.05,0.1,0.5,1,2,5,10,20])
grid_search5 = GridSearchCV(pipeline5, param_grid=params, scoring=neg_rmse, cv=5, n_jobs=4)
a = grid_search5.fit(X_train,y_train)
pickle.dump(a.cv_results_, open('cv_results.g500_and_time.pickle','wb'))

###########################################
####### slp with time
###########################################
pipeline6 = Pipeline(
    [
        ("scale", VarScaler()),
        ("select", VarSelector(varlist=['psl'])),
        ("subset", VarSubset(var="psl", region=[-0.5,3.8,39.7,43.7])), # optimal for tas
        ("weight", VarWeight(var="psl", weight=1)),
        ("adddoy", AddDayOfYear(weight=10)),
        ("knn", KNeighborsRegressor(n_neighbors=15)) # optimal in all experiments so far
    ]
)

params = dict(subset__region = [[-1,4,39,44], [-2,4,38,44], [-5,4,36, 44], [-10, 4, 35,44]],
              adddoy__weight = [0,0.008,0.009, 0.01,0.02,0.03,0.05,0.5,1,5,10,20])
grid_search6 = GridSearchCV(pipeline6, param_grid=params, scoring=neg_rmse, cv=5, n_jobs=4)
a = grid_search6.fit(X_train,y_train)
pickle.dump(a.cv_results_, open('cv_results.slp_and_time.pickle','wb'))

###########################################
####### tas with time
###########################################
pipeline7 = Pipeline(
    [
        ("scale", VarScaler()),
        ("select", VarSelector(varlist=['tas'])),
        ("subset", VarSubset(var="tas", region=[-0.5,3.8,39.7,43.7])), # optimal for tas
        ("weight_a", VarWeight(var="tas", weight=1)),
        ("weight_b", VarWeight(var="psl", weight=1)),
        ("weight_c", VarWeight(var="g500", weight=1)),
        ("adddoy", AddDayOfYear(weight=10)),
        ("knn", KNeighborsRegressor(n_neighbors=15)) # optimal in all experiments so far
    ]
)

params = dict(select__varlist = [['tas'], ['tas','psl'], ['tas','g500'], ['g500','psl'], ['g500','psl','vas'], ['g500','psl','tas']],
              adddoy__weight = [0,0.008,0.009, 0.01,0.02,0.03,0.05,0.5,1,5,10,20])

wgths=[0,0.008,0.009, 0.01,0.02,0.03,0.05,0.5,1,5,10,20]
params = [ { (select__varlist = [['tas'], ['tas','psl'], ['tas','g500'], ['g500','psl'], ['g500','psl','vas'], ['g500','psl','tas']],
              adddoy__weight = [0,0.008,0.009, 0.01,0.02,0.03,0.05,0.5,1,5,10,20])
grid_search7 = GridSearchCV(pipeline7, param_grid=params, scoring=neg_rmse, cv=5, n_jobs=4)
a = grid_search7.fit(X_train,y_train)
pickle.dump(a.cv_results_, open('cv_results.comb_and_time.pickle','wb'))
