import xarray as xr
import numpy as np
import pandas as pd
from tools import fix_coords
from sklearn.neighbors import KNeighborsRegressor

# predictors
era5_dir = '/esarchive/oper/S2S-downscale/era5/'
era5_vars = ['tas','uas','vas','psl','g500','sfcWind']

# tas and sfcWind predictands
uerra_dir = '/esarchive/oper/S2S-downscale/uerra/'
# prlr predictands
mswep_dir = '/esarchive/oper/S2S-downscale/mswep/'

# seas5 predictors
seas5_dir = '/esarchive/oper/S2S-downscale/seas5'

# Load ERA5 predictors
files = [f'{era5_dir}/{var}_1979-2018.nc'
         for var in era5_vars]
era5 = xr.open_mfdataset(files,
                         preprocess = fix_coords,
                         combine="by_coords",
                         chunks = {'lon': 4, 'lat': 4},
                         join = 'left')
# era5 = era5.load(scheduler='threads')
era5 = era5.drop("time_bnds")
#era5 = fix_coords(era5)

# Load UERRA tas (predictand)
files = f'{uerra_dir}/tas_1979-2018.nc'
uerra = xr.open_dataset(files,
                       chunks = {'lon': 4, 'lat': 4},
                       decode_times = True)
# uerra = uerra.load(scheduler='threads')
uerra = uerra.drop("time_bnds")
uerra = fix_coords(uerra)

# Load SEAS5 tas (predictor)
files = f'{seas5_dir}/tas_Nov_DJF-1981-2020.nc'
seas5 = xr.open_dataset(files,
                        chunks = {'ensemble': 5})
seas5 = fix_coords(seas5)

# Add additional coordinates
sdyears = range(1981,2021)
sdmonth = "11"
ltdays = range(30,121)
seas5.coords["sdate"] = [ pd.to_datetime(str(yr) + sdmonth + "01") for yr in sdyears ]
seas5.coords["time"] = pd.to_timedelta([ str(d) + " days" for d in ltdays])
seas5.coords["vdate"] = (["sdate", "time"], seas5.coords["sdate"]+seas5.coords["time"])

# Interp seas5 to era5 grid w nearest
seas5 = seas5.interp_like(era5.tas.isel(time=0, drop=True), method="nearest")

# Region subset
era5 = era5.sel(lon=slice(-1,4), lat=slice(38,44))
seas5 = seas5.sel(lon=slice(-1,4), lat=slice(38,44))
uerra = uerra.sel(lon=slice(0,3.4), lat=slice(40.2,43.2))

# Var subset
era5 = era5.tas
seas5 = seas5.tas
uerra = uerra.tas

# Remove last days (nan in tas)
# NOTE: remove this once fixed
era5 = era5[0:10000,]
uerra = uerra[0:10000,]

#####################
# Leave-One-StartDate-Out Cross-Validation (aka leave 90 days out)
downscaled = None
for looyear in sdyears:
    print(looyear)
    
    # Select SEAS5 forecasts for the given startdate
    X_fcst = seas5.isel(sdate=seas5.sdate.dt.year==looyear)
    
    # Remove all valid dates (eg. all days in DJF 1981/82) from both training sets (X and y)
    X_train = era5.sel(time=np.logical_not(era5.time.isin(X_fcst.vdate)))
    y_train = uerra.sel(time=np.logical_not(uerra.time.isin(X_fcst.vdate)))
    
    # Stack multidimensional arrays for sklearn
    X_train = X_train.stack(feature=("lon","lat"))
    y_train = y_train.stack(pos=("lon","lat"))
    X_fcst = X_fcst.stack(sample=("sdate","time","ensemble")).stack(feature=("lon","lat"))
    #X_train = X_train.to_array().stack(feature=("variable","lon","lat"))
    #y_train = y_train.tas.stack(pos=("lon","lat"))
    #X_fcst = X_fcst.to_array().stack(feature=("variable","lon","lat"))
    
    knn = KNeighborsRegressor(n_neighbors=15)
    y_fcst = knn.fit(X_train,y_train).predict(X_fcst)
    
    # reassign coords and unstack output
    y_fcst = xr.DataArray(y_fcst, dims=["sample","pos"], coords=[X_fcst.coords['sample'], y_train.coords['pos']]).unstack()
    y_fcst.coords["vdate"] = (["sdate", "time"], y_fcst.coords["sdate"]+y_fcst.coords["time"])
    y_fcst.name = y_train.name
    
    if downscaled is None:
        downscaled = y_fcst
    else:
        downscaled = xr.concat([downscaled, y_fcst], dim="sdate")

# Save downscaled data as netcdf
downscaled.transpose("sdate","time","ensemble","lat","lon").to_netcdf("downscaled.nc")

# read the data only after subsetting!
#X_train = X_train.load()
#X_test = X_test.load()
#y_train = y_train.load()

